<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1, user-scalable=1" />
  <title>Flask Livetw</title>

  {% block head %}{% endblock %}

  {live_reload_template_placeholder}
</head>
<body>
  {% block body %}{% endblock %}
</body>
</html>
